import{d as m,ae as d,h as _,j as o,i as v,G as b,k as h,c as r,o as t,l as c,q as k,a as y,F as C,p as l,cW as x,cX as g,m as D,cY as V}from"./index-C_HOOOpe.js";import{u as B}from"./usePageTitle-DzYzWO1v.js";const j=m({__name:"Variables",setup(w){const n=d(),a=()=>{var e;s.value.refresh(),(e=n.value)==null||e.refreshSubscriptions()},u=_(),s=o(()=>v(u.variables.getVariables)),p=o(()=>{var e;return s.value.executed&&((e=s.value.response)==null?void 0:e.length)===0}),i=o(()=>s.value.executed);return B(b.info.variables),(e,F)=>{const f=h("p-layout-default");return t(),r(f,{class:"variables"},{header:c(()=>[D(l(V),{onCreate:a})]),default:c(()=>[i.value?(t(),k(C,{key:0},[p.value?(t(),r(l(x),{key:0,onCreate:a})):(t(),r(l(g),{key:1,ref_key:"table",ref:n,onDelete:a,onUpdate:a},null,512))],64)):y("",!0)]),_:1})}}});export{j as default};
//# sourceMappingURL=Variables-CwiQ3_Um.js.map
