import{d as e,c as o,o as a,p as t,dB as r}from"./index-C_HOOOpe.js";import{u as s}from"./usePageTitle-DzYzWO1v.js";const m=e({__name:"404",setup(n){return s("Page Not Found"),(c,p)=>(a(),o(t(r)))}});export{m as default};
//# sourceMappingURL=404-BmJfYo2k.js.map
