import{d as h,h as x,W as R,u as g,j as i,aI as D,al as I,bt as S,i as T,k as u,c,o as n,l as s,m as B,p as t,a6 as N,a as r,cO as O,cP as m,cQ as P,be as V}from"./index-C_HOOOpe.js";import{u as W}from"./usePageTitle-DzYzWO1v.js";const Q=h({__name:"ConcurrencyLimit",setup(j){const p=x(),y=R("concurrencyLimitId"),d=g(),v=i(()=>[{label:"Details",hidden:D.xl},{label:"Active Task Runs"}]),a=I("tab","Details"),{tabs:_}=S(v,a),b={interval:3e5},L=T(p.concurrencyLimits.getConcurrencyLimit,[y.value],b),e=i(()=>L.response);function f(){d.push(V.concurrencyLimits())}const k=i(()=>e.value?`Concurrency Limit: ${e.value.tag}`:"Concurrency Limit");return W(k),(A,l)=>{const C=u("p-tabs"),w=u("p-layout-well");return n(),c(w,{class:"concurrencyLimit"},{header:s(()=>[e.value?(n(),c(t(P),{key:0,"concurrency-limit":e.value,onDelete:f},null,8,["concurrency-limit"])):r("",!0)]),well:s(()=>[e.value?(n(),c(t(m),{key:0,alternate:"","concurrency-limit":e.value},null,8,["concurrency-limit"])):r("",!0)]),default:s(()=>[B(C,{selected:t(a),"onUpdate:selected":l[0]||(l[0]=o=>N(a)?a.value=o:null),tabs:t(_)},{details:s(()=>[e.value?(n(),c(t(m),{key:0,"concurrency-limit":e.value},null,8,["concurrency-limit"])):r("",!0)]),"active-task-runs":s(()=>{var o;return[(o=e.value)!=null&&o.activeSlots?(n(),c(t(O),{key:0,"active-slots":e.value.activeSlots},null,8,["active-slots"])):r("",!0)]}),_:1},8,["selected","tabs"])]),_:1})}}});export{Q as default};
//# sourceMappingURL=ConcurrencyLimit-B4LauxIM.js.map
