import{d as f,h as x,V as y,W as c,j as r,aC as C,aD as I,I as b,k as h,c as E,a as k,p as e,o as w,l as n,m as s,aE as D,y as F,aF as V}from"./index-C_HOOOpe.js";import{u as B}from"./usePageTitle-DzYzWO1v.js";const j=f({__name:"Event",async setup(g){let a,o;const l=x(),u=y(),v=c("eventDate"),_=c("eventId"),m=r(()=>C(v.value)),p=I({startDate:m,eventId:[_.value]}),t=([a,o]=b(()=>l.events.getFirstEvent(p.value)),a=await a,o(),a),d=r(()=>[{text:"Event Feed",to:u.events()},{text:t.eventLabel}]);return B(`Event: ${t.eventLabel}`),(L,N)=>{const i=h("p-layout-default");return e(t)?(w(),E(i,{key:0,class:"event"},{header:n(()=>[s(e(F),{crumbs:d.value},{actions:n(()=>[s(e(V),{event:e(t)},null,8,["event"])]),_:1},8,["crumbs"])]),default:n(()=>[s(e(D),{event:e(t)},null,8,["event"])]),_:1})):k("",!0)}}});export{j as default};
//# sourceMappingURL=Event-BhorOV-c.js.map
