import{d as m,h as i,i as k,j as t,k as _,c as a,o,l as r,q as f,a as P,F as h,p as s,d2 as w,d3 as y,m as v,d4 as x}from"./index-C_HOOOpe.js";import{u as g}from"./usePageTitle-DzYzWO1v.js";const I=m({__name:"WorkPools",setup(C){const n=i(),l={interval:3e4},e=k(n.workPools.getWorkPools,[{}],l),c=t(()=>e.response??[]),p=t(()=>e.executed&&c.value.length==0),u=t(()=>e.executed);return g("Work Pools"),(b,B)=>{const d=_("p-layout-default");return o(),a(d,{class:"work-pools"},{header:r(()=>[v(s(x))]),default:r(()=>[u.value?(o(),f(h,{key:0},[p.value?(o(),a(s(w),{key:0})):(o(),a(s(y),{key:1,onUpdate:s(e).refresh},null,8,["onUpdate"]))],64)):P("",!0)]),_:1})}}});export{I as default};
//# sourceMappingURL=WorkPools-A1cY_xfi.js.map
