import{j as a,p as c,bw as n}from"./index-C_HOOOpe.js";function i(...t){const r=[...t,"Prefect Server"],o=a(()=>r.map(e=>c(e)).filter(e=>e!==null).join(" • "));n(()=>document.title=o.value)}export{i as u};
//# sourceMappingURL=usePageTitle-DzYzWO1v.js.map
