import{d as I,h as N,W as R,i as S,j as p,ae as T,al as V,G as _,aG as $,k as r,c,o as l,l as e,q as D,a as i,m as s,p as t,aH as E,aI as G,aJ as d,B as H,t as W,aK as v,a6 as j,aL as m,aM as q}from"./index-C_HOOOpe.js";import{u as F}from"./usePageTitle-DzYzWO1v.js";const J={key:0},P=I({__name:"Artifact",setup(K){const k=N(),b=R("artifactId"),w=S(k.artifacts.getArtifact,[b]),a=p(()=>w.response),n=T(!1),y=[{label:"Artifact"},{label:"Details"},{label:"Raw"}],u=V("tab","Artifact"),C=p(()=>a.value?`${_.info.artifact}: ${a.value.key??$(a.value.type)}`:_.info.artifact);return F(C),(L,o)=>{const x=r("p-divider"),g=r("p-button"),A=r("p-content"),B=r("p-tabs"),h=r("p-layout-well");return l(),c(h,{class:"artifact"},{header:e(()=>[a.value?(l(),c(t(q),{key:0,artifact:a.value},null,8,["artifact"])):i("",!0)]),well:e(()=>[a.value?(l(),c(t(m),{key:0,artifact:a.value,alternate:""},null,8,["artifact"])):i("",!0)]),default:e(()=>[a.value?(l(),D("section",J,[s(t(E),{artifact:a.value},null,8,["artifact"]),s(x),t(G).xl?(l(),c(A,{key:0},{default:e(()=>[s(t(d),{artifact:a.value},null,8,["artifact"]),s(g,{class:"artifact__raw-data-button",small:"",onClick:o[0]||(o[0]=f=>n.value=!n.value)},{default:e(()=>[H(W(n.value?"Hide":"Show")+" raw data ",1)]),_:1}),n.value?(l(),c(t(v),{key:0,artifact:a.value},null,8,["artifact"])):i("",!0)]),_:1})):(l(),c(B,{key:1,selected:t(u),"onUpdate:selected":o[1]||(o[1]=f=>j(u)?u.value=f:null),tabs:y},{artifact:e(()=>[s(t(d),{artifact:a.value},null,8,["artifact"])]),details:e(()=>[s(t(m),{artifact:a.value},null,8,["artifact"])]),raw:e(()=>[s(t(v),{artifact:a.value},null,8,["artifact"])]),_:1},8,["selected"]))])):i("",!0)]),_:1})}}});export{P as default};
//# sourceMappingURL=Artifact-kXQOfeRq.js.map
